/*
 *  Ipd version 1.0, Copyright (C) 1995 G. Hutchings
 *  Ipd comes with ABSOLUTELY NO WARRANTY.
 *  This is free software, and you are welcome to redistribute it
 *  under certain conditions; see the file COPYING for details.
 */

#ifndef lint
static char rcsid[] = "$Id: ipd.c,v 1.6 1995/06/13 09:10:43 glen Exp $";
#endif

/* The main program */

#ifdef CURSES
#include <curses.h>
#include <signal.h>
#else
#include <stdio.h>
#endif

#include <string.h>
#include <stdarg.h>
#include "config.h"
#include "rule.h"
#include "proto.h"

/* Program version */
#define VERSION		"1.0"

/* Contest types */
#define COMBINATION	1
#define ROUND_ROBIN	2
#define EVOLUTION	3
#define TERRITORIAL	4

/* Various defaults */
#define TURNS		100
#define GENERATIONS	100
#define POPULATION	100
#define INTERVAL	1
#define WIDTH		60
#define HEIGHT		20

/* Rule match flags */
#define NO_MATCH	-1
#define AMBIGUOUS	-2

char *ofile = NULL;		/* Output filename */
char *prog;			/* My name */
char buf[BUFSIZ];		/* Scribble buffer */
char errbuf[BUFSIZ];		/* Error stream buffer */
int height = HEIGHT;		/* Territory height (-t) */
int num_entries = 0;		/* No. of rules in contest */
int num_gen = GENERATIONS;	/* No. of generations (-e and -t) */
int num_rules = 0;		/* Total no. of rules */
int num_turns = TURNS;		/* No. of turns to iterate for */
int plot_output = 0;		/* Whether to plot output (-e) */
int population = POPULATION;	/* Population size (-e) */
int print_interval = INTERVAL;	/* Output interval (-e and -t) */
int rule_weight = 1;		/* Default command line rule weight */
int show_moves = 0;		/* Whether to output moves */
int show_rules = 0;		/* Whether to list rules */
int show_syms = 0;		/* Whether to show symbols */
int show_usage = 0;		/* Whether to print usage */
int type = COMBINATION;		/* Type of contest */
int width = WIDTH;		/* Territory width (-t) */

struct entry *entries;		/* The list of contest entries */

#ifdef CURSES
int screen = 0;			/* Whether to do screen display */
int screen_init = 0;		/* Is screen initialised? */
#endif

int
main(int argc, char *argv[])
{
    extern char *optarg;
    extern int optind;
    int c, i, num;
    FILE *fp;

    /* What's my name? */
    prog = argv[0];

    /* Count no. of rules */
    while (NAME(num_rules) != NULL)
	num_rules++;

    /* Allocate space for entries */
    if ((entries = ALLOC(struct entry, num_rules)) == NULL)
	error("out of memory");

    /* Parse arguments */
    while ((c = getopt(argc, argv, "cretmR:w:f:n:g:i:p:x:y:o:PSslhv")) != EOF) {
	switch (c) {
	case 'c':
	    type = COMBINATION;
	    break;
	case 'r':
	    type = ROUND_ROBIN;
	    break;
	case 'e':
	    type = EVOLUTION;
	    break;
	case 't':
	    type = TERRITORIAL;
	    break;
	case 'm':
	    show_moves = 1;
	    break;
	case 'R':
	    switch ((num = rule_match(optarg, 0))) {
	    case NO_MATCH:
		error("no rule matches `%s'", optarg);
		break;
	    case AMBIGUOUS:
		error("more than one rule matched by `%s'", optarg);
		break;
	    default:
		add_entry(num, rule_weight);
		break;
	    }

	    break;
	case 'w':
	    rule_weight = atoi(optarg);
	    if (rule_weight < 1)
		error("%s: invalid rule weight", optarg);
	    break;
	case 'f':
	    if (!strcmp(optarg, "-")) {
		read_entries(stdin);
	    } else {
		if ((fp = fopen(optarg, "r")) == NULL)
		    error("%s: can't open rule file", optarg);
		read_entries(fp);
		fclose(fp);
	    }

	    break;
	case 'n':
	    num_turns = atoi(optarg);
	    if (num_turns < 1)
		error("%s: invalid no. of turns", optarg);
	    break;
	case 'g':
	    num_gen = atoi(optarg);
	    if (num_gen < 1)
		error("%s: invalid no. of generations", optarg);
	    break;
	case 'p':
	    population = atoi(optarg);
	    if (population < 10)
		error("%s: invalid population", optarg);
	    break;
	case 'i':
	    print_interval = atoi(optarg);
	    if (print_interval < 0)
		error("%s: invalid print interval", optarg);
	    break;
	case 'x':
	    width = atoi(optarg);
	    if (width < 2)
		error("%s: invalid territory width", optarg);
	    break;
	case 'y':
	    height = atoi(optarg);
	    if (height < 2)
		error("%s: invalid territory height", optarg);
	    break;
	case 'o':
	    ofile = optarg;
	    break;
	case 'P':
	    plot_output = 1;
	    break;
	case 'S':
#ifdef CURSES
	    screen = 1;
	    break;
#else
	    fprintf(stderr, "Sorry, no screen display available.\n");
	    return 1;
#endif
	case 's':
	    show_syms = 1;
	    break;
	case 'l':
	    show_rules = 1;
	    break;
	case 'h':
	    show_usage = 1;
	    break;
	case 'v':
	    printf("Ipd version %s, (c) Glenn Hutchings 1995\n", VERSION);
	    return 0;
	default:
	    fprintf(stderr, "Type `%s -h' for help\n", prog);
	    exit(1);
	}
    }

    /* Show usage if required */
    if (show_usage)
	usage();

    /* Show rules if required */
    if (show_rules || show_syms) {
	for (i = 0; i < num_rules; i++) {
	    if (show_syms)
		printf("%c\t", SYM(i));
	    printf("%s\n", NAME(i));
	}

	return 0;
    }

    /* Remaining arguments (if any) are rule files */
    while (optind < argc) {
	if (!strcmp(argv[optind], "-")) {
	    read_entries(stdin);
	} else {
	    if ((fp = fopen(argv[optind], "r")) == NULL)
		error("%s: can't open rule file", argv[optind]);
	    read_entries(fp);
	    fclose(fp);
	}

	optind++;
    }

    /* Check there's something to do */
    if (num_entries == 0)
	error("no contest entries specified");

    /* Check for incompatible options */
#ifdef CURSES
    if (screen && ofile != NULL)
	error("can't use -S and -o together");
    if (screen && show_moves)
	error("can't use -S and -m together");
    if (screen && type != EVOLUTION && type != TERRITORIAL)
	error("-S only works with -e and -t");
#endif

    /* Redirect output if required */
    if (ofile != NULL)
	if (freopen(ofile, "w", stdout) == NULL)
	    error("can't open %s", ofile);

    /* Initialise random stream */
    IRAND(SEED);

    /* Initialise display if required */
#ifdef CURSES
    if (screen) {
	screen_init = 1;
	initscr();
	noecho();
	signal(SIGINT, quit);
	if (type == EVOLUTION && num_entries > LINES - 1)
	    error("screen not tall enough for %d entries", num_entries);
	if (type == TERRITORIAL && width > COLS - 1)
	    error("screen not wide enough for width %d", width);
	if (type == TERRITORIAL && height > LINES - 4)
	    error("screen not tall enough for height %d", height);
    }
#endif

    /* Buffer error output if required */
    if (show_moves)
	setbuf(stderr, errbuf);

    /* Run the contest */
    switch (type) {
    case COMBINATION:
	do_combination();
	break;
    case ROUND_ROBIN:
	do_round_robin();
	break;
    case EVOLUTION:
	do_evolution();
	break;
    case TERRITORIAL:
	do_territorial();
	break;
    }

    /* Close display if required */
#ifdef CURSES
    if (screen)
	end_screen();
#endif

    return 0;
}

void
do_combination()
{
    static char buf[BUFSIZ];
    int i, j, s1, s2;

    for (i = 0; i < num_entries; i++) {
	for (j = i; j < num_entries; j++) {
	    dilemma(entries[i].rule, entries[j].rule, &s1, &s2);
	    sprintf(buf, "%s vs %s", NAME(entries[i].rule),
		    NAME(entries[j].rule));
	    printf("%-50s  %d %d\n", buf, s1, s2);
	}
    }
}

void
do_round_robin()
{
    int i, j, s1, s2;

    /* Run the contest */
    for (i = 0; i < num_entries; i++) {
	for (j = i; j < num_entries; j++) {
	    dilemma(entries[i].rule, entries[j].rule, &s1, &s2);
	    entries[i].score += s1;
	    entries[i].games++;
	    entries[j].score += s2;
	    entries[j].games++;
	}
    }

    /* Average the scores */
    for (i = 0; i < num_entries; i++)
	entries[i].score /= entries[i].games;

    /* Sort the results */
    qsort(entries, num_entries, sizeof(struct entry), scorecmp);

    /* Print them out */
    for (i = 0; i < num_entries; i++)
	printf("%-30s  %.1f\n", NAME(entries[i].rule), entries[i].score);
}

void
do_evolution()
{
    int i, j, *pop, s1, s2, g, p;
    double tot, pfac;

    /* If plotting, unbuffer stdout for immediate results */
    if (plot_output)
	setbuf(stdout, NULL);

    /* Allocate population list */
    if ((pop = ALLOCA(int, population)) == NULL)
	error("out of stack space");

    /* Set initial dummy scores */
    for (i = 0; i < num_entries; i++)
	entries[i].score = 1.0;

    /* Run generations */
    for (g = 0; g < num_gen; g++) {
	/* Allocate population according to scores */
	for (tot = 0.0, i = 0; i < num_entries; i++)
	    tot += entries[i].num * entries[i].score;

	for (p = 0, i = 0; i < num_entries; i++) {
	    pfac = entries[i].num * entries[i].score;
	    entries[i].num = (int) population * pfac / tot;

	    p += entries[i].num;
	}

	/*
	 * Population may have slipped a bit due to rounding
	 * errors - boost it up again
	 */
	while (p++ < population)
	    entries[random_entry()].num++;

	/* Print initial population */
	if (g == 0)
	    print_population(0);

	/* Set up population list */
	for (p = 0, i = 0; i < num_entries; i++)
	    for (j = 0; j < entries[i].num; j++)
		pop[p++] = i;

	/* Initialise next generation */
	for (i = 0; i < num_entries; i++)
	    entries[i].score = entries[i].games = 0;

	/* Do interactions */
	for (i = 0; i < population; i++) {
	    for (j = i; j < population; j++) {
		dilemma(entries[pop[i]].rule, entries[pop[j]].rule, &s1, &s2);
		entries[pop[i]].score += s1;
		entries[pop[i]].games++;
		entries[pop[j]].score += s2;
		entries[pop[j]].games++;
	    }
	}

	/* Normalize scores */
	for (i = 0; i < num_entries; i++)
	    if (entries[i].games > 0)
		entries[i].score /= entries[i].games;

	/* Print population if required */
	if ((g + 1) % print_interval == 0)
	    print_population(g + 1);
    }

    /* Print final population */
    print_population(g);
}

void
do_territorial()
{
    int i, j, i2, j2, k, p = 0, g, s1, s2, best, max, change;
    static int dir[4][2] = {1, 0, 0, 1, -1, 0, 0, -1};
    struct loc **board;

    /* Allocate board space */
    if ((board = ALLOCA(struct loc *, width)) == NULL)
	error("out of stack space");
    for (i = 0; i < width; i++)
	if ((board[i] = ALLOCA(struct loc, height)) == NULL)
	    error("out of stack space");

    /* Initialise with random entries */
    for (i = 0; i < width; i++)
	for (j = 0; j < height; j++)
	    board[i][j].rule[p] = entries[random_entry()].rule;

    /* Print initial board */
    print_territory(board, p, 0);

    /* Run generations */
    for (g = 0; g < num_gen; g++) {
	/* Initialise generation */
	change = 0;
	for (i = 0; i < width; i++) {
	    for (j = 0; j < height; j++) {
		board[i][j].score = 0;
		board[i][j].played = 0;
	    }
	}

	/* Play each cell against its neighbours */
	for (i = 0; i < width; i++) {
	    for (j = 0; j < height; j++) {
		board[i][j].played = 1;
		for (k = 0; k < 4; k++) {
		    i2 = i + dir[k][0];
		    if (i2 == width)
			i2 = 0;
		    if (i2 == -1)
			i2 = width - 1;
		    j2 = j + dir[k][1];
		    if (j2 == height)
			j2 = 0;
		    if (j2 == -1)
			j2 = height - 1;
		    if (board[i2][j2].played)
			continue;
		    dilemma(board[i][j].rule[p],
			    board[i2][j2].rule[p],
			    &s1, &s2);
		    board[i][j].score += s1;
		    board[i2][j2].score += s2;
		}
	    }
	}

	/* Check each cell for rule changes */
	for (i = 0; i < width; i++) {
	    for (j = 0; j < height; j++) {
		board[i][j].rule[1 - p] = board[i][j].rule[p];
		max = 0;
		for (k = 0; k < 4; k++) {
		    i2 = i + dir[k][0];
		    if (i2 == width)
			i2 = 0;
		    if (i2 == -1)
			i2 = width - 1;
		    j2 = j + dir[k][1];
		    if (j2 == height)
			j2 = 0;
		    if (j2 == -1)
			j2 = height - 1;
		    if (board[i2][j2].score > max) {
			max = board[i2][j2].score;
			best = board[i2][j2].rule[p];
		    }
		}

		if (max > board[i][j].score) {
		    board[i][j].rule[1 - p] = best;
		    change = 1;
		}
	    }
	}

	/* End if no changes */
	if (!change) {
#ifdef CURSES
	    if (!screen)
#endif
		printf("\nTerritory is now static\n");
	    return;
	}

	/* Flip states to new generation */
	p = 1 - p;

	/* Print board if required */
	if (print_interval &&
	    (g + 1) % print_interval == 0 &&
	    (g + 1) != num_gen) {
	    print_territory(board, p, g + 1);
	}
    }

    /* Print final board */
    print_territory(board, p, num_gen);
}

/* Iterate a Prisoner's Dilemma between two rules */
void
dilemma(int r1, int r2, int *s1, int *s2)
{
    int i, turn, *last1, *last2, m1, m2, p1, p2;
    int (*f1) () = Rules[r1].func;
    int (*f2) () = Rules[r2].func;

    /* Allocate move history */
    if ((last1 = ALLOCA(int, num_turns)) == NULL ||
	(last2 = ALLOCA(int, num_turns)) == NULL)
	error("out of stack space");

    /* Initialise scores */
    *s1 = *s2 = 0;

    /* Iterate the dilemma */
    for (turn = 0; turn < num_turns; turn++) {
	/* Get moves */
	m1 = last1[turn] = (*f1) (turn, last1, last2);
	m2 = last2[turn] = (*f2) (turn, last2, last1);

	/* Calculate points earned */
	if (m1 == COOPERATE && m2 == COOPERATE)
	    p1 = p2 = REWARD;
	else if (m1 == COOPERATE && m2 == DEFECT)
	    p1 = SUCKER, p2 = TEMPTATION;
	else if (m1 == DEFECT && m2 == COOPERATE)
	    p1 = TEMPTATION, p2 = SUCKER;
	else
	    p1 = p2 = PUNISHMENT;

	/* Update scores */
	*s1 += p1;
	*s2 += p2;
    }

    /* Print moves to stderr if requested */
    if (show_moves) {
	fprintf(stderr, "\n%25s :", NAME(r1));
	for (i = 0; i < num_turns; i++)
	    fprintf(stderr, "%c", (last1[i] == COOPERATE ? 'C' : 'D'));
	fprintf(stderr, "\n%25s :", NAME(r2));
	for (i = 0; i < num_turns; i++)
	    fprintf(stderr, "%c", (last2[i] == COOPERATE ? 'C' : 'D'));
	fprintf(stderr, "\n");
    }
}

/* Read rule entries */
void
read_entries(FILE * fp)
{
    char buf[BUFSIZ], *name, *num;
    int n, w, line = 0;

    while (fgets(buf, BUFSIZ, fp) != NULL) {
	line++;

	/* Skip blank and comment lines */
	if (!strlen(buf) || buf[0] == '#')
	    continue;

	/* Get rule name */
	name = strtok(buf, "\t\n");
	if (name == NULL)
	    continue;
	if ((n = rule_match(name, 1)) == NO_MATCH)
	    error("line %d: `%s': unknown rule", line, name);

	/* Get weighting */
	num = strtok(NULL, "\t\n");
	w = (num == NULL ? 1 : atoi(num));
	if (w < 1)
	    error("line %d: %s: invalid weighting", line, num);

	/* Add entry */
	add_entry(n, w);
    }
}

/* Add a rule to the entry list */
void
add_entry(int rule, int num)
{
    int i, e;

    for (e = -1, i = 0; i < num_entries; i++)
	if (entries[i].rule == rule)
	    e = i;

    if (e < 0) {
	e = num_entries++;
	entries[e].rule = rule;
	entries[e].num = num;
	entries[e].score = 0.0;
	entries[e].games = 0;
    } else {
	entries[e].num += num;
    }
}

/* Return the number of a matched rule name */
int
rule_match(char *str, int exact)
{
    int match = NO_MATCH, len, i;
    char *cp;

    len = strlen(str);
    for (i = 0; i < num_rules; i++) {
	if (!strcasecmp(str, NAME(i)))
	    return i;
	if (exact)
	    continue;
	for (cp = NAME(i); *cp != '\0'; cp++) {
	    if (!strncasecmp(str, cp, len)) {
		if (match == NO_MATCH)
		    match = i;
		else
		    match = AMBIGUOUS;
	    }
	}
    }

    return match;
}

/* Return a random rule entry according to weightings */
int
random_entry()
{
    int i, tot = 0, total = 0, weight;

    /* Find total weight */
    for (i = 0; i < num_entries; i++)
	total += entries[i].num;

    /* Choose weighted random entry */
    weight = RANDOM(total);
    for (i = 0; i < num_entries; i++) {
	tot += entries[i].num;
	if (tot > weight)
	    return i;
    }

    error("serious internal problem");
    /* NOTREACHED */
}

/* Print current population */
void
print_population(int gen)
{
    int i, j, gap, n, line;

#ifdef CURSES
    if (screen) {
	erase();
	sprintf(buf, " Generation: %d ", gen);
	standout();
	mvaddstr(0, (COLS - strlen(buf)) / 2, buf);
	standend();
	gap = (LINES - 2) / num_entries;
	line = (LINES + gap - gap * num_entries) / 2;

	for (i = 0; i < num_entries; i++) {
	    n = (int) (53.0 * entries[i].num / population + 0.5);
	    sprintf(buf, "%25s ", NAME(entries[i].rule));
	    mvaddstr(line + i * gap, 0, buf);
	    standout();
	    for (j = 0; j < n; j++)
		mvaddch(line + i * gap, 27 + j, ' ');
	    standend();
	}

	move(0, 0);
	refresh();
	sleep(1);
	return;
    }
#endif

    if (plot_output) {
	printf("%3d", gen);
	for (i = 0; i < num_entries; i++)
	    printf(" %3d", entries[i].num);
	printf("\n");
    } else {
	printf("\nGeneration %d:\n", gen);
	for (i = 0; i < num_entries; i++)
	    printf("  %-25s  %d\n", NAME(entries[i].rule), entries[i].num);
    }
}

/* Print a territory */
void
print_territory(struct loc ** t, int state, int gen)
{
    int i, j, line, col;
    char c;

    /* Print title */
#ifdef CURSES
    if (screen) {
	col = (COLS - width) / 2;
	line = (LINES - height) / 2;
	erase();
	sprintf(buf, " Generation: %d ", gen);
	standout();
	mvaddstr(line - 2, (COLS - strlen(buf)) / 2, buf);
	standend();
    } else {
	printf("\nGeneration: %d\n\n", gen);
    }
#else
    printf("\nGeneration: %d\n\n", gen);
#endif

    /* Print territory */
    for (j = 0; j < height; j++) {
	for (i = 0; i < width; i++) {
	    c = SYM(t[i][j].rule[state]);
#ifdef CURSES
	    if (screen)
		mvaddch(line + j, col + i, c);
	    else
		putchar(c);
#else
	    putchar(c);
#endif
	}

#ifdef CURSES
	if (!screen)
#endif
	    putchar('\n');
    }

#ifdef CURSES
    if (screen) {
	move(0, 0);
	refresh();
	sleep(1);
    }
#endif
}

/* Score comparison function */
int
scorecmp(struct entry * e1, struct entry * e2)
{
    if (e1->score > e2->score)
	return -1;
    else if (e1->score < e2->score)
	return 1;

    return 0;
}

/* Close down display */
void
end_screen()
{
#ifdef CURSES
    if (screen_init) {
	echo();
	mvcur(0, COLS - 1, LINES - 1, 0);
	endwin();
    }
#endif
}

/* Deal with a Control C */
void
quit()
{
    end_screen();
    exit(0);
}

/* Give an error message and exit */
void
error(char *fmt,...)
{
    va_list ap;

    va_start(ap, fmt);
    vsprintf(buf, fmt, ap);
    va_end(ap);

#ifdef CURSES
    if (screen)
	end_screen();
#endif

    fprintf(stderr, "%s: Error: %s\n", prog, buf);
    exit(1);
}

/* Give a usage message and exit */
void
usage()
{
    printf("Usage: %s [options] [files]\n", prog);

    printf("\nEntering rules:\n");
    printf("  -R <rule>   enter <rule> in the contest\n");
    printf("  -w <num>    set weighting for -R rules\n");
    printf("  -f <file>   enter rules from <file>\n");

    printf("\nContest types:\n");
    printf("  -c          run rule combinations (default)\n");
    printf("  -r          run a round-robin contest\n");
    printf("  -e          run an evolution contest\n");
    printf("  -t          run a territorial contest\n");

    printf("\nControlling output:\n");
    printf("  -o <file>   write output to <file>\n");
    printf("  -n <num>    set no. of turns (default %d)\n", TURNS);
    printf("  -g <num>    set no. of generations (default %d)\n", GENERATIONS);
    printf("  -p <num>    set population size (default %d)\n", POPULATION);
    printf("  -i <num>    set printing interval (default %d)\n", INTERVAL);
    printf("  -x <num>    set territory width (default %d)\n", WIDTH);
    printf("  -y <num>    set territory height (default %d)\n", HEIGHT);
    printf("  -P          plot population levels\n");
    printf("  -m          write moves to stderr\n");
    printf("  -S          enable screen output\n");

    printf("\nMiscellaneous:\n");
    printf("  -l          write rule list\n");
    printf("  -s          write rules and their symbols\n");
    printf("  -v          write the program version\n");
    printf("  -h          this usage message\n");

    exit(0);
}
