.TH IPD 1 
.fi
.SH NAME
ipd
\-
iterated Prisoner's Dilemma
.SH SYNOPSIS
.na
.nf
ipd [options] [files]
.SH DESCRIPTION
.fi
.B ipd
runs a contest of the iterated Prisoner's Dilemma, using a
selection of built-in decision rules.

.B ipd
is based on the computer tournament run by Robert Axelrod and
described in his book \fBThe Evolution of Co-operation\fP.
Most of the decision rules included here are re-coded (with
varying degrees of success) from this source.
.SH OPTIONS
.na
.nf
.fi

.SS "Entering rules"
.TP 8
.B "-R <rule>"
Add \fBrule\fP to the list of entries in the contest.  You will
have to quote the rule name if it contains spaces, to avoid it
being hassled by the shell.  A rule name entered using this
option does not have to exactly match a rule, as long as
enough matches to make it unambiguous.
.TP
.B "-w <num"
Set the rule weighting for subsequent rules specified using
the \fB-R\fP option.  Default is 1.
.TP
.B "-f <file>"
Add the rules contained in \fBfile\fP to the list of entries.
See \fBRULE FILES\fP, below, for the format.  Unlike \fB-R\fP,
rule names listed in a rule file must exactly match rules.
You can use the \fB-l\fP option to get a list of existing
rules (and even pipe the list back into \fBipd\fP).

Rule files may also be listed after all the options.  In this
case, the special name "-" means to read from stdin.

.SS "Contest types"
.TP 8
.B "-c"
Set Combination mode (the default).  See \fBCONTEST TYPES\fP,
below.
.TP
.B "-r"
Set Round-Robin mode.  See \fBCONTEST TYPES\fP, below.
.TP
.B "-e"
Set Evolution mode.  See \fBCONTEST TYPES\fP, below.
.TP
.B "-t"
Set Territorial mode.  See \fBCONTEST TYPES\fP, below.

.SS "Controlling output"
.TP 8
.B "-o <file>"
Write output to \fBfile\fP, instead of stdout.
.TP
.B "-n <num>"
Set the number of turns in each iterated dilemma (default
100).
.TP
.B "-g <num>"
Set the number of generations in the evolution or territorial
contests (default 100).
.TP
.B "-p <num>"
Set the population size for the evolution contest (default
100).
.TP
.B "-i <num>"
Set the output interval for the evolution and territorial
contests.  This displays the current contest state every
\fBnum\fP generations.  Default is 1.
.TP
.B "-x <num>"
.TP
.B "-y <num>"
Sets the width and height, respectively, of the grid for the
territorial contest.  Defaults are 60 and 20.
.TP
.B "-P"
Plot the population levels in the evolution contest in a
table, instead of the default text output.  This output is
suitable for feeding to a plotting program such as
.B gnuplot.
The first column is the generation number, and each following
column is the population level for each rule.  Rules are in
the same order as in the rule file.
.TP
.B "-m"
Print the moves made in each game to stderr.  Use wisely: this
can produce a LOT of output!
.TP
.B "-S"
Enable
.MS curses 3
screen display.  This is only available for the evolution and
territorial contests.  Type Control-C to quit.

.SS "Miscellaneous"
.TP
.B "-l"
Print a list of available decision rules to stdout, and exit.
This is a good way of producing a default file for use with
the \fB-f\fP option.
.TP
.B "-s"
Print a list of the decision rules and their symbols, as a key
to the display in the territorial contest.
.TP
.B "-v"
Print the version of \fBipd\fP.
.TP
.B "-h"
Print an option summary.
.SH THE PRISONER'S DILEMMA
.na
.nf
.fi
The Prisoner's Dilemma, for those who don't already know, is a
two-player, two-option game.  The options available to each
player are `co-operate' (be nice) or `defect' (be nasty).
Played with the standard payoffs, possible results are:

1) Both co-operate, and get 3 points each.

2) One co-operates and the other defects.  The co-operator
gets no points (the sucker's payoff) while the defector gets a
juicy 5 points.

3) Both defect, and get only 1 point each.

Cold, calculating logic says the best move is to always defect
-- you always get more than you would have done by
co-operating.  The dilemma is that when your opponent deduces
the same, you're both going to get only 1 point each.  But if
only you had both co-operated, you'd both be 2 points richer!

The main point of Robert Axelrod's book is that even though in
a one-shot Prisoner's Dilemma defection is the only logical
outcome, when the players play the game repeatedly those
prepared to be nicer will do better, because there is the
possibility of building up mutual trust.  Axelrod ran a series
of contests which pitted different playing strategies against
each other, with some pretty interesting results.  \fBipd\fP
allows you to run your own versions of these contests.

The books mentioned in the \fBSEE ALSO\fP section are highly
recommended if you want to know more about the Prisoner's
Dilemma (especially the first one).
.SH RULE FILES
.na
.nf
.fi
Rules can be entered in a contest via the \fB-R\fP option (on
the command line) or via the \fB-f\fP option, from a file.  In
a rules file, each rule appears on a line on its own.  The
name of the rule must be exactly as it appears in the output
from the \fB-l\fP option (although case is ignored).  You may
insert blank lines and comment lines (those beginning with #)
wherever you like.

Each rule may be followed on its line by a `weighting' number,
separated from the rule name by one or more tabs.  (NOTE:
spaces are \fInot\fP allowed between a rule name and its
weighting, since spaces are valid in rule names.)  This
weighting number has different meanings depending on the type
of contest.  A rule with no weighting number specified has a
weighting of 1.

If a rule is entered in the contest more than once, its
weightings are added together.
.SH CONTEST TYPES
.na
.nf
.fi

.SS "Combination"
This is the simplest type of contest.  Each rule is paired
against each other rule, and the individual results are
printed to stdout.  Rule weightings are ignored.

.SS "Round-Robin"
Each rule is paired against each other rule, and the results
for each rule are averaged and sorted decreasing score order.
Rule weightings are ignored.

.SS "Evolution"
An initial population of players is set up, each using
randomly-allocated rules according to their weighting.  In
each generation every player interacts with every other
player, and the average scores gained by each rule is
calculated.  The popularity of each rule in the next
generation is proportional to its current popularity and the
average score gained.

.SS "Territorial"
A number of Prisoner's Dilemma players occupy a rectangular
grid, each using one of the decision rules.  In each
generation, each player interacts only with their four
immediate neighbours (the grid wraps around in both
directions).  After the interactions, each player has gained a
total score.  If any of a player's neighbours has gained a
better score than they have, that player changes their rule to
that of the neighbour with the best score.  And so, after many
generations, the distribution of rules will change.

The initial distribution of rules in the grid is chosen
randomly, according to the weighting given to each rule.
If the situation occurs that no player decides to change
rules, \fBipd\fP will stop the contest.
.SH RESTRICTIONS
.na
.nf
.fi
All the decision rules are hard-coded.  If you want to add
more, you have to recompile.
.SH BUGS
.fi
.fi
None.  But there may be a few undocumented features. 8-)
.SH SEE ALSO
.fi
.nf
\fBThe Evolution of Co-operation\fP, by Robert Axelrod.
\fBPrisoner's Dilemma\fP, by William Poundstone.
\fBThe Selfish Gene\fP (2nd Edition), by Richard Dawkins.
.SH AUTHOR
.na
.nf
Glenn Hutchings (zondo@hunting2.demon.co.uk)
