#!%%PERLPATH%%

# Run ipd evolution contest and produce a Gnuplot file

# $Id: ipdplot.pl,v 1.3 1995/06/13 08:54:36 glen Exp $

# Parse arguments
if (&getopts('n:p:g:q')) {
    fprintf STDERR "Usage: ipdplot [-q] [-n num] [-g gen] [-p pop] rulefile\n";
    exit 1;
}

# Get rule filename
die "No rule file specified\n" if (@ARGV == 0);
$rulefile = $ARGV[0];
$datafile = $rulefile . ".dat";
$plotfile = $rulefile . ".gp";

# Set defaults
$opt_n = 40 if !defined $opt_n;
$opt_g = 10 if !defined $opt_g;
$opt_p = 50 if !defined $opt_p;

# Read rules file
open (RULES, "<" . $rulefile) || die "Can't open $rulefile: $!\n";
while (<RULES>) {
    push(@rules, $1) if /^([^#\n][^\t\n]*)/;
}
close RULES;

# Invoke ipd
$cmd = "ipd -eP -n$opt_n -p$opt_p -g$opt_g -o $datafile $rulefile";
print "Running ${cmd}\n" if !defined $opt_q;
if (system "$cmd") {
    unlink $datafile;
    die "ipdplot: aborted\n";
}

# Create Gnuplot file
open (GP, ">" . $plotfile) || die "Can't open $plotfile: $!\n";
&plot("# Ipd evolution plot file (produced by ipdplot).");
&plot("set data style lines");
&plot("set grid");
&plot("set key");
&plot("set xlabel 'Generation'");
&plot("set ylabel 'Population'");
&plot("plot \\");
for ($i = 0; $i < @rules; $i++) {
    $col = $i+2;
    $sep = ($i < @rules-1 ? ", \\" : "");
    &plot("  '$datafile' using 1:${col} title '$rules[$i]'${sep}");
}
close GP;

# Er... that's it
exit 0;

# Send a line to the Gnuplot file
sub plot {
    local ($line) = @_;
    printf GP "%s\n", $line;
}

# Own version of getopts (lousy error checking on the library one)
sub getopts {
    local ($argdefs) = @_;
    local (@args, $_, $first, $rest);
    local ($errs) = 0;
    local ($[) = 0;

    @args = split(/ */, $argdefs);
    while (@ARGV && ($_ = $ARGV[0]) =~ /^-(.)(.*)/) {
	($first, $rest) = ($1, $2);
	$pos = index($argdefs, $first);
	if ($pos >= $[) {
	    if ($args[$pos+1] eq ':') {
		shift(@ARGV);
		$rest = shift(@ARGV) if ($rest eq '');
		if ($rest ne '' || $rest =~ /^-/) {
		    eval "\$opt_$first = \$rest;";
		} else {
		    print STDERR "Option requires argument: -$first\n";
		    ++$errs;
		}
	    } else {
		eval "\$opt_$first = 1";
		if ($rest eq '') {
		    shift(@ARGV);
		} else {
		    $ARGV[0] = "-$rest";
		}
	    }
	} else {
	    print STDERR "Unknown option: -$first\n";
	    ++$errs;
	    if ($rest ne '') {
		$ARGV[0] = "-$rest";
	    } else {
		shift(@ARGV);
	    }
	}
    }

    return $errs;
}
